@@ -18,19 +18,24 @@ from argparse import ArgumentParser, RawDescriptionHelpFormatter
 from contextlib import suppress
 from datetime import datetime
 from errno import ENOENT
+import json
 import logging
 import logging.config
 import os
 from pathlib import Path
 import re
 import sys
-from urllib.parse import urlparse
 
 from dateutil.parser import parse as datetimeparse
 from dateutil.relativedelta import relativedelta
 from dateutil.tz import tzlocal
 import jmespath
 import jsonschema
+try:
+    from referencing import Registry, Resource
+    from referencing.jsonschema import DRAFT202012
+except ImportError:
+    pass  # python < 3.8
 from ruamel.yaml import YAML
 from ruamel.yaml.constructor import ConstructorError
 
@@ -244,7 +249,7 @@ class DataProcessor:
             },
             VARIABLES_KEY: {
                 'patternProperties': {
-                    r'^[A-z_]\w*$': {'type': 'string'},
+                    r'^[A-z_]\w*$': {'description': 'any valid variable name'},
                 },
                 'additionalProperties': False,
                 'type': 'object',
@@ -281,6 +286,8 @@ class DataProcessor:
         r"_FORMAT_(?P<name>\w+)",
         re.M | re.S)
 
+    UNBOUND_ORIGINAL = 'YP_ORIGINAL'
+
     def __init__(self):
         self.is_process_include = True
         self.is_process_variable = True
@@ -318,6 +325,7 @@ class DataProcessor:
                 # Ignore the MERGE flag.
                 data, parent_filenames, variable_map = self.load_include_file(
                     data, parent_filenames, variable_map)[0:3]
+                root = data
             type_of_data = type(data)
             items_iter = None
             skip_keys = set()
@@ -434,6 +442,9 @@ class DataProcessor:
             is_merge = (self.MERGE_KEY in orig_value)
             include_filename = self.process_variable(
                 value[self.INCLUDE_KEY])
+            if self.VARIABLES_KEY in value:
+                for key, val in value[self.VARIABLES_KEY].copy().items():
+                    value[self.VARIABLES_KEY][key] = self.process_variable(val)
             try:
                 loaded_value = self.include_dict[include_filename]
                 filename = include_filename
@@ -504,7 +515,7 @@ class DataProcessor:
         item: object,
         variable_map: dict = None,
     ) -> object:
-        """Substitute environment variables into a string value.
+        """Substitute (environment) variables into a string value.
 
         Return `item` as-is if not `.is_process_variable` or if `item` is not a
         string.
@@ -516,14 +527,18 @@ class DataProcessor:
         `.unbound_placeholder` is None, raise an `UnboundVariableError`.
 
         If `NAME` is not defined in the `.variable_map` and
-        `.unbound_placeholder` is not None, substitute `NAME` with the value
+        `.unbound_placeholder` equals to the value of
+        `DataProcessor.UNBOUND_ORIGINAL`, then leave the original syntax
+        unchanged.
+
+        If `NAME` is not defined in the `.variable_map` and
+        `.unbound_placeholder` is not `None`, substitute `NAME` with the value
         of `.unbound_placeholder`.
 
         :param item: Item to process. Do nothing if not a str.
         :return: Processed item on success.
         :param variable_map: :py:attr:`.variable_map` in the local scope,
                              may have additional variables.
-
         """
         if not self.is_process_variable or not isinstance(item, str):
             return item
@@ -538,19 +553,22 @@ class DataProcessor:
             match = self.REC_SUBSTITUTE.match(tail)
             if match:
                 groups = match.groupdict()
-                substitute = groups['symbol']
                 if len(groups['escape']) % 2 == 0:
                     if groups['name'] in variable_map:
                         substitute = variable_map[groups['name']]
                     elif groups['name'].startswith('YP_TIME'):
                         substitute = self._process_time_variable(
                             groups['name'])
+                    elif self.unbound_placeholder == self.UNBOUND_ORIGINAL:
+                        substitute = groups['symbol']
                     elif self.unbound_placeholder is not None:
                         substitute = str(self.unbound_placeholder)
                     else:
                         raise UnboundVariableError(groups['name'])
-                if groups['cast']:
-                    if groups['head'] or tail != item:
+                else:
+                    substitute = groups['symbol']
+                if substitute != groups['symbol'] and groups['cast']:
+                    if groups['head'] or groups['tail'] or tail != item:
                         raise ValueError(
                             f'{item}: bad substitution expression')
                     try:
@@ -575,6 +593,9 @@ class DataProcessor:
                             f'{item}: bad substitution value: {substitute}')
                     ret = substitute
                     tail = ''
+                elif not any(groups[k] for k in ('head', 'escape', 'tail')):
+                    ret = substitute
+                    tail = ''
                 else:
                     ret += (
                         groups['head']
@@ -684,21 +705,42 @@ class DataProcessor:
         """
         if not schema_location:
             return
-        schema = {"$ref": schema_location}
-        if not urlparse(schema_location).scheme:
-            schema_path = Path(schema_location)
-            if schema_path.exists():
-                schema = {"$ref": schema_path.absolute().as_uri()}
-            elif self.schema_prefix:
-                schema = {"$ref": self.schema_prefix + schema_location}
         try:
-            jsonschema.validate(schema=schema, instance=data)
+            registry = Registry(retrieve=self.get_schema_file)
+            jsonschema.Draft202012Validator(
+                {
+                    '$schema': 'https://json-schema.org/draft/2020-12/schema',
+                    '$ref': schema_location,
+                },
+                registry=registry,
+            ).validate(data)
+        except NameError:
+            ref = 'file://' + str(self._get_schema_file(schema_location))
+            jsonschema.validate(schema={'$ref': ref}, instance=data)
         except jsonschema.exceptions.ValidationError as exc:
             logging.error(f'not ok {out_file_name}')
             logging.exception(exc)
             raise
-        else:
-            logging.info(f'ok {out_file_name}')
+        logging.info(f'ok {out_file_name}')
+
+    def get_schema_file(self, schema_location: str):
+        schema_path = self._get_schema_file(schema_location)
+        return Resource(
+            contents=json.loads(schema_path.read_text()),
+            specification=DRAFT202012)
+
+    def _get_schema_file(self, schema_location: str):
+        if schema_location.startswith('file://'):
+            schema_location = schema_location.replace('file://', '', 1)
+        schema_path = Path(schema_location)
+        if not schema_path.exists():
+            for prefix in [self.schema_prefix]:  # FIXME
+                if prefix.startswith('file://'):
+                    prefix = prefix.replace('file://', '', 1)
+                if (Path(prefix) / schema_path).exists():
+                    schema_path = Path(prefix) / schema_path
+                    break
+        return schema_path
 
 
 def main(argv=None):
@@ -749,7 +791,12 @@ def main(argv=None):
         '--unbound-placeholder',
         metavar='VALUE',
         default=None,
-        help='Substitute an unbound variable with VALUE instead of failing')
+        help=(
+            'Substitute an unbound variable with VALUE instead of failing.'
+            f' Use {DataProcessor.UNBOUND_ORIGINAL} as VALUE to leave the'
+            ' original syntax unchanged on unbound variables.'
+        ),
+    )
     parser.add_argument(
         '--no-process-include',
         dest='is_process_include',
